(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DP_Transformation
imports
  QDP_Framework
  Lexicographic_Extension
begin

fun
  initial_dpp :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs  \<Rightarrow> ('f, 'v) dpp"
where
  "initial_dpp shp Q R = (DP shp (useful_rules Q R), {}, Q, {}, R)"


(* @{term "Tinf"} characterizes termination. *)
lemma Tinf_SN_conv: "(Tinf R = {}) = (SN R)"
proof
  assume "Tinf R = {}" thus "SN R" using not_SN_imp_Tinf by blast
next
  assume "SN R" thus "Tinf R = {}" unfolding Tinf_def SN_defs by blast
qed


lemma Tinf_lemma_1:
  assumes "wwf_qtrs Q R"
    and "Fun f ts \<in> Tinf (qrstep Q R)" (is "?t \<in> _")
  shows "\<exists>l r u \<sigma>.
  (l, r) \<in> useful_rules Q R \<and> 
  r \<unrhd> u \<and> 
  is_Fun u \<and> 
  (?t, l \<cdot> \<sigma>) \<in> (nrqrstep Q R)^* \<and> 
  (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep Q R \<and> 
  r \<cdot> \<sigma> \<unrhd> u \<cdot> \<sigma> \<and>
  l \<cdot> \<sigma> \<in> Tinf (qrstep Q R) \<and>
  u \<cdot> \<sigma> \<in> Tinf (qrstep Q R) \<and>
  \<not> (l \<rhd> u)"
proof -
  let ?U = "useful_rules Q R"
  let ?qr = "qrstep Q R"
  let ?rqr = "rqrstep Q R"
  let ?nrqr = "nrqrstep Q R"
  from assms have "\<And>s. ?t \<rhd> s \<Longrightarrow> SN_on ?qr {s}"
    and "\<not> SN_on ?qr {?t}" unfolding Tinf_def by auto
  from `\<not> SN_on ?qr {?t}` have "\<exists>A. A 0 = ?t \<and> (\<forall>i. (A i,A(Suc i)) \<in> ?qr)"
    unfolding SN_on_def by auto
  then obtain "A" where "A 0 = ?t" and "\<forall>i. (A i,A(Suc i)) \<in> ?qr" by auto
  hence "!!i. (A i,A(Suc i)) \<in> (?rqr \<union> ?nrqr)" by (auto simp: qrstep_iff_rqrstep_or_nrqrstep)
  hence "(A 0,A 1) \<in> ?rqr \<or> (A 0,A 1) \<in> ?nrqr" by force
  hence "\<exists>i. (A 0,A i) \<in> ?nrqr^* \<and> (A i,A(Suc i)) \<in> ?rqr"
  proof
    assume A: "(A 0,A 1) \<in> ?rqr"
    have "(A 0,A 0) \<in> ?nrqr^*" by auto
    with A show ?thesis by auto
  next
    assume B: "(A 0,A 1) \<in> ?nrqr"
    from `?t \<in> Tinf ?qr ` have "SN_on ?nrqr {?t}"
      using SN_on_subterms_imp_SN_on_nrqrstep unfolding Tinf_def by blast
    from `A 0 = ?t`[symmetric] `SN_on ?nrqr {?t}` have "SN_on ?nrqr {A 0}" by simp
    from `!!i. (A i,A(Suc i)) \<in> (?rqr \<union> ?nrqr)` `SN_on ?nrqr {A 0}`
    have "\<exists>i. (A i,A(Suc i)) \<in> ?rqr \<and> (\<forall>j<i. (A j,A(Suc j)) \<in> ?nrqr
      \<and> \<not> (A j,A(Suc j)) \<in> ?rqr)"
      using union_iseq_SN_on_imp_first_step[where R = ?rqr and S = ?nrqr and t = A] by force
    then obtain i where first_rstep: "(A i,A(Suc i)) \<in> ?rqr"
      and "!!j. j<i \<Longrightarrow> (A j,A(Suc j)) \<in> ?nrqr \<and> \<not> (A j,A(Suc j)) \<in> ?rqr" by auto
    hence "\<forall>j<i. (A j,A(Suc j)) \<in> ?nrqr" by auto
    hence "(A 0,A i) \<in> ?nrqr^*" by (induct i) auto
    with first_rstep show ?thesis by auto
  qed
  then obtain i where "(A 0,A i) \<in> ?nrqr^*" and step: "(A i,A(Suc i)) \<in> ?rqr" by auto
  then obtain l r \<sigma>
    where NF: "\<forall>v\<lhd>l \<cdot> \<sigma>. v \<in> NF_terms Q"
    and "(l, r) \<in> R" and "l \<cdot> \<sigma> = A i" and "r \<cdot> \<sigma> = A (Suc i)" by auto
  with step have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?rqr" by auto
  from `(l, r) \<in> R` only_useful_rules[OF NF] have "(l, r) \<in> ?U"
    unfolding useful_rules_def by auto
  from `A 0 = ?t` and `(A 0,A i) \<in> ?nrqr^*` have "(?t,A i) \<in> ?nrqr^*" by simp
  hence "\<exists>ls. A i = Fun f ls" by (rule nrqrsteps_preserve_root_fun)
  then obtain ls' where "A i = Fun f ls'" by auto
  have A: "\<And>i. (ts ! i, ls' ! i) \<in> ?qr^*" (is "\<And>i. ?P i")
  proof -
    fix j
    from `(?t, A i) \<in> ?nrqr^*` have "(args ?t!j, args (Fun f ls')!j) \<in> ?qr^*"
      unfolding `A i = Fun f ls'` by (rule nrqrsteps_imp_arg_qrsteps)
    thus "?P j" by simp
  qed
  from `(?t,A i) \<in> ?nrqr^*` and `A i = Fun f ls'` have "(?t,Fun f ls') \<in> ?nrqr^*" by simp
  from `l\<cdot>\<sigma> = A i` and `A i = Fun f ls'` have "l\<cdot>\<sigma> = Fun f ls'" by simp
  with `(?t,Fun f ls') \<in> ?nrqr^*` have "(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*" by simp
  from `(?t,Fun f ls') \<in> ?nrqr^*` have numargs: "num_args ?t = num_args (Fun f ls')"
    by (rule nrqrsteps_num_args)
  from `wwf_qtrs Q R` and `(l,r) \<in> R` obtain g ls where "l = Fun g ls"
    using wwf_qtrs_imp_left_fun by best
  hence "l\<cdot>\<sigma> = Fun g (map (\<lambda>t. t\<cdot>\<sigma>) ls)" by simp
  with `l\<cdot>\<sigma> = Fun f ls'` have "g = f" and "ls' = map (\<lambda>t. t\<cdot>\<sigma>) ls" by auto
  hence B: "\<forall>i<length ls'. ls' ! i = (ls ! i) \<cdot> \<sigma>" by auto
  have "\<forall>i < num_args ?t. (ts ! i, (ls ! i) \<cdot> \<sigma>) \<in> ?qr^*"
  proof (intro allI impI)
    fix i assume "i < num_args (Fun f ts)"
    with numargs have "i < length ls'" by simp
    with A[of i] show "(ts ! i, ls ! i \<cdot> \<sigma>) \<in> ?qr^*" using B by simp
  qed
  from `l \<cdot> \<sigma> = Fun f ls'` numargs have "num_args ?t = num_args (l\<cdot>\<sigma>)"  by simp
  with `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>`
  have "!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i" by auto
  have "!!i. i < num_args ?t \<Longrightarrow> ?t \<rhd> (ts!i)"
  proof -
    fix i assume "i < num_args ?t"
    hence "ts!i \<in> set ts" by simp
    thus "?t \<rhd> (ts!i)" by (rule supt.arg)
  qed
  with `!!s. ?t \<rhd> s \<Longrightarrow> SN_on ?qr {s}` have "!!i. i < num_args ?t \<Longrightarrow> SN_on ?qr {ts ! i}" by auto
  with `\<forall>i<num_args ?t. (ts!i,(ls!i)\<cdot>\<sigma>) \<in> ?qr^*` 
  have "!!i. i < num_args ?t \<Longrightarrow> SN_on ?qr {(ls!i) \<cdot> \<sigma>}" using steps_preserve_SN_on[of _ _ ?qr] by blast
  with `!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i` and `num_args ?t = num_args (l\<cdot>\<sigma>)`
    and `l\<cdot>\<sigma> = Fun f ls'` and `!!i. i < num_args ?t \<Longrightarrow> SN_on ?qr {(ls!i) \<cdot> \<sigma>}`
  have argsSN: "!!i. i < num_args (Fun f ls') \<Longrightarrow> SN_on ?qr {ls' ! i}" by auto
  have "\<forall>s\<lhd>Fun f ls'. SN_on ?qr {s}" 
  proof (intro allI impI)
    fix s
    assume tmp: "Fun f ls' \<rhd> s"
    then obtain C where "C \<noteq> \<box>" and "Fun f ls' = C\<langle>s\<rangle>" by auto
    then obtain bef aft C where id: "Fun f ls' = Fun f (bef @ C\<langle>s\<rangle> # aft)" by (cases C, auto)
    hence len: "length bef < num_args (Fun f ls')" by auto
    with argsSN[OF len] id have "SN_on ?qr {C \<langle>s\<rangle>}" by auto
    with restrict_SN_imp_SN_subt[OF ctxt_closed_qrstep _ ctxt_imp_supteq[of C s]]
    show "SN_on ?qr {s}" by auto
  qed
  hence "\<forall>s\<lhd>l\<cdot>\<sigma>. SN_on ?qr {s}" unfolding `l\<cdot>\<sigma> = Fun f ls'` by simp 
  let ?B = "\<lambda>j. A (i + j)"
  have "?B 0 = l\<cdot>\<sigma>" unfolding `l\<cdot>\<sigma> = A i` by simp
  have "\<forall>i. (?B i,?B(Suc i)) \<in> ?qr" using `\<forall>i. (A i,A(Suc i)) \<in> ?qr` by simp
  with `?B 0 = l\<cdot>\<sigma>` have "\<not> SN_on ?qr {l \<cdot> \<sigma>}" unfolding SN_on_def by best
  with `\<forall>s\<lhd>l\<cdot>\<sigma>. SN_on ?qr {s}` have tinf_ls: "l \<cdot> \<sigma> \<in> Tinf ?qr" unfolding Tinf_def by simp
  have "\<And> x. x \<in> vars_term r \<Longrightarrow> l \<rhd> Var x"
  proof -
    fix x
    assume "x \<in> vars_term r" hence "x \<in> vars_term l" using `(l,r) \<in> ?U` and `wwf_qtrs Q R` unfolding wwf_qtrs_def wwf_rule_def useful_rules_def by auto
    thus "l \<rhd> Var x" using `l = Fun g ls` supteq_Var[OF `x \<in> vars_term l`]
      unfolding supt_supteq_conv by auto
  qed
  hence "\<And> x. x \<in> vars_term r \<Longrightarrow> l \<cdot> \<sigma> \<rhd> Var x \<cdot> \<sigma>"
    using supt_stable[of l _ \<sigma>] unfolding supt_supteq_conv by best
  hence xSN: "\<And> x. x \<in> vars_term r \<Longrightarrow> SN_on ?qr {Var x \<cdot> \<sigma>}" using tinf_ls unfolding Tinf_def by auto
  let ?B = "\<lambda>j. A (Suc i + j)"
  from `r\<cdot>\<sigma> = A (Suc i)` have "?B 0 = r\<cdot>\<sigma>" by simp
  from `\<forall>i. (A i,A(Suc i)) \<in> ?qr` have "\<forall>i. (?B i,?B(Suc i)) \<in> ?qr" by auto
  with `?B 0 = r\<cdot>\<sigma>` have "\<not> SN_on ?qr {r \<cdot> \<sigma>}" unfolding SN_on_def by best
  then obtain t' where "r\<cdot>\<sigma> \<unrhd> t'" and "t' \<in> Tinf ?qr" using not_SN_imp_subt_Tinf by auto
  have "\<forall>x \<in> vars_term r. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t')"
  proof
    fix x assume "x \<in> vars_term r"
    hence "SN_on ?qr {Var x \<cdot> \<sigma>}" using xSN by simp
    show "\<not> Var x\<cdot>\<sigma> \<unrhd> t'"
    proof
      assume "Var x\<cdot>\<sigma> \<unrhd> t'"
      with `SN_on ?qr {Var x \<cdot> \<sigma>}` have "SN_on ?qr {t'}" using restrict_SN_imp_SN_subt[OF ctxt_closed_qrstep] by auto
      with `t' \<in> Tinf ?qr` show "False" unfolding Tinf_def by auto
    qed
  qed
  with `r\<cdot>\<sigma> \<unrhd> t'` have "\<exists>u. r \<unrhd> u \<and> t' = u\<cdot>\<sigma>" using subt_instance_and_not_subst_imp_subt[where s = r and \<sigma> = \<sigma>] by auto
  then obtain u where "r \<unrhd> u" and "t' = u\<cdot>\<sigma>" by best
  have "is_Fun u"
  proof (rule ccontr)
    assume "\<not> is_Fun u"
    hence "\<exists>x. u = Var x" by (cases u) simp_all
    then obtain x where "u = Var x" by best
    with `r \<unrhd> u` have "r \<unrhd> Var x" by auto
    hence "x \<in> vars_term r" using var_subt_imp_subt_in_vars by best
    with xSN `u = Var x` have "SN_on ?qr {u \<cdot> \<sigma>}" by simp
    with `t' = u\<cdot>\<sigma>` have "SN_on ?qr {t'}" by simp
    with `t' \<in> Tinf ?qr` show "False" unfolding Tinf_def by simp
  qed
  from `t' = u\<cdot>\<sigma>` and `t' \<in> Tinf ?qr` have "u \<cdot> \<sigma> \<in> Tinf ?qr" by simp
  have "\<not> l \<rhd> u"
  proof
    assume "l \<rhd> u"
    hence "l \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>" by (rule supt_stable)
    with `l \<cdot> \<sigma> \<in> Tinf ?qr` have "SN_on ?qr {u \<cdot> \<sigma>}" unfolding Tinf_def by best
    with `u \<cdot> \<sigma> \<in> Tinf ?qr` show "False" unfolding Tinf_def by auto
  qed
  from `r \<unrhd> u` have "r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>)" by (rule supteq_stable)
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `is_Fun u` and `(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*` 
    `(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr` and `u \<cdot> \<sigma> \<in> Tinf ?qr` and `r \<cdot> \<sigma> \<unrhd> u \<cdot> \<sigma>` and `l \<cdot> \<sigma> \<in> Tinf ?qr`
    and `\<not> l \<rhd> u`
  show ?thesis by blast
qed

lemma SN_on_Var:
   assumes "wwf_qtrs Q R" shows "SN_on (qrstep Q R) {Var x}" (is "SN_on _ {?x}")
proof -
  let ?qr = "qrstep Q R"
  show "SN_on ?qr {?x}"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain A where "A 0 = ?x" and rsteps: "\<forall>i. (A i,A(Suc i)) \<in> ?qr"
      unfolding SN_on_def by best
    hence "(?x,A 1) \<in> ?qr" using spec[OF rsteps, of 0] by auto
    then obtain l r C \<sigma> where "(l,r) \<in> R" and x: "Var x = C\<langle>l \<cdot> \<sigma>\<rangle>" by auto
    with wwf_qtrs_imp_left_fun and assms obtain f ls where "l = Fun f ls" by blast
    with x obtain ls where "Var x = C\<langle>Fun f ls\<rangle>" by auto
    thus False by (cases C, auto)
  qed
qed

lemma Tinf_var:
  assumes "wwf_qtrs Q R" shows "Var x \<notin> Tinf (qrstep Q R)"
  using SN_on_Var[OF assms] unfolding Tinf_def by auto
 
lemma Tinf_lemma:
  assumes wwf: "wwf_qtrs Q R" and tinf: "t\<in> Tinf (qrstep Q R)"
  shows "\<exists>l r u \<sigma>.
    (l, r) \<in> useful_rules Q R \<and> r \<unrhd> u \<and> is_Fun u \<and> (t, l \<cdot> \<sigma>) \<in> (nrqrstep Q R)^*
    \<and> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep Q R \<and> r \<cdot> \<sigma> \<unrhd> u \<cdot> \<sigma> \<and> l \<cdot> \<sigma> \<in> Tinf (qrstep Q R)
    \<and> u \<cdot> \<sigma> \<in> Tinf (qrstep Q R) \<and> \<not> (l \<rhd> u)"
proof (cases t)
  case (Var x)
  with Tinf_var[OF wwf] tinf have False by auto
  thus ?thesis ..
next
  case (Fun f ts)
  from Tinf_lemma_1[OF wwf tinf[simplified Fun]]
  show ?thesis unfolding Fun .
qed

lemma Tinf_imp_def_root_qrstep:
  assumes wwf: "wwf_qtrs Q R" and tinf: "t \<in> Tinf (qrstep Q R)"
  shows "\<exists>f ss. t = Fun f ss \<and> defined (useful_rules Q R) (f, length ss)"
proof -
  from Tinf_var[OF wwf] tinf obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  with tinf have tinf: "Fun f ts \<in> Tinf (qrstep Q R)" by auto
  from Tinf_lemma[OF wwf tinf]
  obtain l r \<sigma> where lr: "(l,r) \<in> useful_rules Q R" and "(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*" 
    by blast
  with wwf[unfolded wwf_qtrs_wf_trs wf_trs_def]
  obtain g ls where l: "l = Fun g ls" by blast
  with nrqrsteps_preserve_root[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*`]
    nrqrsteps_num_args[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*`]
  have "g = f" and "length ls = length ts" by auto
  with lr l have "defined (useful_rules Q R) (f, length ts)" unfolding defined_def by auto
  with t show ?thesis by auto
qed

lemma Tinf_imp_Fun: 
   assumes "wwf_qtrs Q R" and "t \<in> Tinf (qrstep Q R)"
   shows "\<exists>f ts. t = Fun f ts"
proof (cases t)
  fix x assume "t = Var x"
  from `wwf_qtrs Q R` have "SN_on (qrstep Q R) {t}"
    unfolding `t = Var x` using SN_on_Var by best
  with `t \<in> Tinf (qrstep Q R)` show ?thesis unfolding Tinf_def by auto
next
  fix f ts assume "t = Fun f ts" thus ?thesis by auto
qed


lemma nrqrstep_imp_sharp_qrstep:
  assumes step: "(s,t) \<in> nrqrstep Q R"
  shows "(sharp_term shp s, sharp_term shp t) \<in> qrstep Q R"
proof -
  let ?qr = "qrstep Q R"
  let ?nr = "nrqrstep Q R"
  let ?Q = "Q"
  let ?R = "R"
  let ?Iqr = "qrstep ?Q ?R"
  from step obtain l r C \<sigma> where C: "C \<noteq> \<box>" and  lr: "(l,r) \<in> R" and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" and NF: "\<forall>u\<lhd>l \<cdot> \<sigma>.  u \<in> NF_terms Q" 
    unfolding nrqrstep_def rstep_r_c_s_def by auto
  from C obtain D f bef aft where C: "C = More f bef D aft" by (cases C, auto)
  let ?Dl = "D\<langle>l \<cdot> \<sigma>\<rangle>"
  let ?Dr = "D\<langle>r \<cdot> \<sigma>\<rangle>"
  let ?C = "More (shp f) bef \<box> aft"
  from s t C have s: "s = Fun f (bef @ ?Dl # aft)" and t: "t = Fun f (bef @ ?Dr # aft)" by auto
  from lr NF have "(?Dl,?Dr) \<in> ?qr" by auto
  hence "(?C\<langle>?Dl\<rangle>,?C\<langle>?Dr\<rangle>) \<in> ?Iqr" by (rule qrstep_closed_ctxt )
  with s t show ?thesis by auto
qed

lemma nrqrsteps_imp_sharp_qrsteps:
  assumes step: "(s,t) \<in> (nrqrstep Q R)^*"
  shows "(sharp_term shp s, sharp_term shp t) \<in> (qrstep Q R)^*"
using step
proof (induct rule: rtrancl_induct, simp)
  case (step u v)
  from nrqrstep_imp_sharp_qrstep[OF step(2), of shp] step(3)
  show ?case by auto
qed

lemma Tinf_lemma_DP:
  assumes "wwf_qtrs Q R" "s \<in> Tinf (qrstep Q R)"
  shows "\<exists>t u. t \<in> Tinf (qrstep Q R) \<and> u \<in> Tinf (qrstep Q R) \<and>
    (sharp_term shp s, sharp_term shp t) \<in> (qrstep Q R)^* \<and>
    (sharp_term shp t, sharp_term shp u) \<in> CS (DP shp R)"
proof -
  obtain l r u \<sigma> where "(l, r) \<in> useful_rules Q R" and "r \<unrhd> u" and "is_Fun u"
    and nrseq: "(s, l \<cdot> \<sigma>) \<in> (nrqrstep Q R)^*"
    and "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep Q R" and "l \<cdot> \<sigma> \<in> Tinf (qrstep Q R)"
    and "u \<cdot> \<sigma> \<in> Tinf (qrstep Q R)"
    and "\<not> (l \<rhd> u)"
    using Tinf_lemma[OF assms] by auto
  let ?l = "sharp_term shp l"
  let ?r = "sharp_term shp u"
  from `is_Fun u` obtain f us where "u = Fun f us" by (cases u) auto
  hence "u\<cdot>\<sigma> = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) us)" (is "_ = Fun f ?us'") by simp
  hence "root u = root (u\<cdot>\<sigma>)" unfolding `u = Fun f us` by simp
  from `wwf_qtrs Q R` and `u \<cdot> \<sigma> \<in> Tinf (qrstep Q R)` obtain h vs where "u \<cdot> \<sigma> = Fun h vs"
    and "defined (useful_rules Q R) (h, length vs)"
    using Tinf_imp_def_root_qrstep by best
  from nrseq have "(sharp_term shp s, sharp_term shp (l\<cdot>\<sigma>)) \<in> (qrstep Q R)^*"
    by (rule nrqrsteps_imp_sharp_qrsteps)
  from `(l, r) \<in> useful_rules Q R` and `wwf_qtrs Q R` obtain g ts
    where "l = Fun g ts" unfolding wwf_qtrs_def useful_rules_def
    using useful_rules_subset by (cases l) blast+
  from `u\<cdot>\<sigma> = Fun h vs` and `u = Fun f us` have "f = h" and "length us = length vs" by auto
  hence "defined (useful_rules Q R) (f, length us)"
    using `defined (useful_rules Q R) (h, length vs)` by simp
  from useful_rules_subset and `(l, r) \<in> useful_rules Q R` have "(l, r) \<in> R" by auto
  from `defined (useful_rules Q R) (f, length us)` have "defined R (f, length us)"
    unfolding useful_rules_def useful_rule_def defined_def by auto
  from `(l, r) \<in> R` and `r \<unrhd> u` and `defined R (f, length us)` and `\<not> l \<rhd> u`
    have "(?l, ?r) \<in> DP shp R" unfolding DP_def `u = Fun f us` by auto
  hence "(?l\<cdot>\<sigma>, ?r\<cdot>\<sigma>) \<in> CS (DP shp R)" by auto
  hence "(sharp_term shp (l\<cdot>\<sigma>), sharp_term shp (u\<cdot>\<sigma>)) \<in> CS (DP shp R)"
    unfolding `l = Fun g ts` `u = Fun f us` sharp_term_subst_distr by simp
  with `l \<cdot> \<sigma> \<in> Tinf (qrstep Q R)` and `u \<cdot> \<sigma> \<in> Tinf (qrstep Q R)`
    and `(sharp_term shp s, sharp_term shp (l\<cdot>\<sigma>)) \<in> (qrstep Q R)^*`
    show ?thesis by auto
qed

lemma Tinf_lemma_DP_explicit_subst':
  assumes "wwf_qtrs Q R" and "s \<in> Tinf (qrstep Q R)"
  shows "\<exists>l r u \<sigma>. (l, r) \<in> DP shp R \<and> r \<cdot> \<sigma> = sharp_term shp u \<and> u \<in> Tinf (qrstep Q R)
    \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (qrstep Q R)^*"
proof -
  from Tinf_lemma_DP[OF assms, of shp] obtain t u
    where t: "t \<in> Tinf (qrstep Q R)" and u: "u \<in> Tinf (qrstep Q R)"
    and steps: "(sharp_term shp s, sharp_term shp t) \<in> (qrstep Q R)^*"
    and step: "(sharp_term shp t, sharp_term shp u) \<in> CS (DP shp R)" by auto
  from step obtain l r \<sigma> where lr: "(l,r) \<in> DP shp R" and tl: "sharp_term shp t = l \<cdot> \<sigma>" and ur: "sharp_term shp u = r \<cdot> \<sigma>" 
    by (induct, auto)
  show ?thesis
    by (rule exI[of _ l], rule exI[of _ r], rule exI[of _ u], rule exI[of _ \<sigma>], insert steps lr tl ur t u, auto)
qed

definition wf_qtrs :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "wf_qtrs Q R \<equiv> wwf_qtrs Q R \<and> (\<forall>t\<in>Q. is_Fun t)"

lemma wf_qtrs_wf_trs_conv: "wf_qtrs {} = wf_trs"
proof (rule ext)
  fix R
  show "wf_qtrs {} R = wf_trs R"
    unfolding wf_qtrs_def wf_trs_def wwf_qtrs_def
    by (force simp: useful_rule_empty)
qed

lemma Tinf_lemma_DP_explicit_subst:
  assumes wf: "wf_qtrs Q R" and s: "s \<in> Tinf (qrstep Q R)"
    and ndef:
    "\<And>f n ss. defined (useful_rules Q R) (f, n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  shows "\<exists>l r u \<sigma>. (l,r) \<in> DP shp (useful_rules Q R) \<and> r \<cdot> \<sigma> = sharp_term shp u
    \<and> u \<in> Tinf (qrstep Q R) \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (qrstep Q R)^*
    \<and> l \<cdot> \<sigma> \<in> NF_terms Q"
proof -
  let ?qr = "qrstep Q R"
  let ?U = "useful_rules Q R"
  let ?nrqr = "nrqrstep Q R"
  let ?rqr = "rqrstep Q R"
  from wf have wf: "wwf_qtrs Q R" and wfQ: "\<forall>t\<in>Q. is_Fun t"
    unfolding wf_qtrs_def by auto
  let ?Q = "Q" and ?R = "R"
  let ?Iqr = "qrstep ?Q ?R"
  obtain l r u \<sigma> where lr: "(l,r) \<in> ?U" and "r \<unrhd> u" and "is_Fun u" and "(s,l\<cdot>\<sigma>) \<in> ?nrqr^*"
    and step: "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr" and "l \<cdot> \<sigma> \<in> Tinf ?qr" and "u \<cdot> \<sigma> \<in> Tinf ?qr" and "\<not> l \<rhd> u"
    using Tinf_lemma[OF wf s] by auto
  let ?u = "u\<cdot>\<sigma>"
  let ?l = "sharp_term shp l"
  let ?r = "sharp_term shp u"
  let ?\<sigma> = \<sigma>
  obtain f us where "u = Fun f us" using `is_Fun u` by (cases u) auto
  have "root u = root ?u" unfolding `u = Fun f us` by simp
  obtain h vs where "?u = Fun h vs" and "defined ?U (h, length vs)"
    using Tinf_imp_def_root_qrstep[OF wf `?u \<in> Tinf ?qr`] by auto
  from `(s,l\<cdot>\<sigma>) \<in> ?nrqr^*` have "(sharp_term shp s, sharp_term shp (l \<cdot> \<sigma>)) \<in> ?Iqr^*"
    by (rule nrqrsteps_imp_sharp_qrsteps)
  from `(l,r) \<in> ?U` and wf[unfolded wwf_qtrs_wf_trs] obtain g ts where l: "l = Fun g ts" unfolding wf_trs_def by blast
  have "sharp_term shp (l\<cdot>\<sigma>) = ?l\<cdot>?\<sigma>" unfolding `l = Fun g ts` sharp_term_subst_distr by simp
  with `(sharp_term shp s,sharp_term shp (l\<cdot>\<sigma>)) \<in> ?Iqr^*` have "(sharp_term shp s,?l\<cdot>?\<sigma>) \<in> ?Iqr^*" by simp
  from `(u\<cdot>\<sigma>) = Fun h vs` and `u = Fun f us` have "f = h" and "length us = length vs" by auto
  with `defined ?U (h, length vs)` have "defined ?U (f, length us)" by simp
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `defined ?U (f, length us)` and `\<not>(l \<rhd> u)` have "(?l,?r) \<in> DP shp ?U" unfolding DP_def `u = Fun f us` by auto
  have "?r\<cdot>?\<sigma> = sharp_term shp ?u" unfolding `u = Fun f us` sharp_term_subst_distr by simp
  from step have NF: "\<forall>u\<lhd>l\<cdot>\<sigma>. u \<in> NF_terms Q" unfolding rqrstep_def rstep_r_c_s_def by auto
  from lr l have gdef: "defined ?U (g, length ts)" unfolding defined_def by auto
  note ndef = ndef[OF gdef]
  note NF = NF[unfolded l]
  have "?l \<cdot> ?\<sigma> \<in> NF_terms ?Q"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    from not_NF_termsE[OF this]
    obtain q C \<gamma> where q: "q \<in> ?Q" and match: "?l \<cdot> ?\<sigma> = C\<langle>q \<cdot> \<gamma>\<rangle>" .
    from q have qg: "q \<cdot> \<gamma> \<notin> NF_terms Q" unfolding NF_def by force
    show False
    proof (cases C)
      case (More ff bef D aft)
      hence "Fun g ts \<cdot> ?\<sigma> \<rhd> q \<cdot> \<gamma>" using match l by auto
      with NF have "q \<cdot> \<gamma> \<in> NF_terms Q" by auto
      with qg show False by simp
    next
      case Hole
      with match have match: "?l\<cdot>?\<sigma> = q \<cdot> \<gamma>" by auto
      show ?thesis 
      proof (cases q)
        case (Fun ff qq)
        from match[unfolded l this] have ff: "ff = shp g" and qq: "length ts = length qq" using map_eq_imp_length_eq[of _ ts _ qq] by auto
        from ndef[OF qq[symmetric]] q[unfolded Fun ff]  
        show ?thesis by auto
      next
        case (Var x)
        from q[unfolded Var] wfQ show False by auto
      qed
    qed 
  qed
  hence "?l \<cdot> ?\<sigma> \<in> NF_terms ?Q" by auto
  with `(?l,?r) \<in> DP shp ?U` and `?r\<cdot>?\<sigma> = sharp_term shp ?u`
    and `?u \<in> Tinf ?qr` and `(sharp_term shp s, ?l\<cdot>?\<sigma>) \<in> ?Iqr^*`
  show ?thesis by blast
qed
  

lemma someI_ex4:
  "\<exists>w x y z. P (w, x, y, z) \<Longrightarrow> P (SOME (w, x, y, z). P (w, x, y, z))"
by (simp add: some_eq_ex)

fun
  next4 :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow>
    ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)subst"
where
  "next4 shp Q R s = (SOME (l, r, u, \<sigma>).
    (l, r) \<in> DP shp (useful_rules Q R) \<and> r \<cdot> \<sigma> = sharp_term shp u \<and> u \<in> Tinf (qrstep Q R)
    \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (qrstep Q R)^*
    \<and> l \<cdot> \<sigma> \<in> NF_terms Q)"

fun
  concrete_seq :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term \<Rightarrow> nat \<Rightarrow>
    ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) subst"
where
  "concrete_seq shp Q R s 0 = (s, next4 shp Q R s)" |
  "concrete_seq shp Q R r (Suc n) = (
    let ( _, _, _, s, _) = concrete_seq shp Q R r n in
    (s, next4 shp Q R s))"


theorem not_SN_imp_min_ichain_both:
  assumes "wf_qtrs Q R" and "\<not> SN (qrstep Q R)" 
  and ndef: "\<And> f n ss. defined (useful_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  shows "\<exists>s t \<sigma>. ichain (initial_dpp shp Q R) s t \<sigma> \<and> ((\<forall> f n. defined (useful_rules Q R) (f,n) \<longrightarrow> \<not> defined (useful_rules Q R) (shp f,n)) \<longrightarrow> min_ichain (initial_dpp shp Q R) s t \<sigma>)"
proof -
  let ?qr = "qrstep Q R"
  let ?Q = "Q" and ?R = "R"
  let ?Iqr = "qrstep ?Q ?R"
  from `\<not>(SN(qrstep Q R))` obtain s' where "\<not> (SN_on ?qr {s'})"
    unfolding SN_on_def by auto
  then obtain s where "s' \<unrhd>  s" and "s \<in> Tinf ?qr" using not_SN_imp_subt_Tinf by auto
  from assms have "wf_qtrs Q R" by auto
  let ?I = "\<lambda>i. fst(concrete_seq shp Q R s i)"
  let ?S = "\<lambda>i. fst(snd(concrete_seq shp Q R s i))"
  let ?T = "\<lambda>i. fst(snd(snd(concrete_seq shp Q R s i)))"
  let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq shp Q R s i))))"
  let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP shp (useful_rules Q R) \<and> r\<cdot>\<sigma> = sharp_term shp u
    \<and> u \<in> Tinf ?qr \<and> (sharp_term shp s,l\<cdot>\<sigma>) \<in> ?Iqr^* \<and> l \<cdot> \<sigma> \<in> NF_terms ?Q"
  let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
  have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof
    fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
    proof (induct i)
      case 0 
      have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)"
        using Tinf_lemma_DP_explicit_subst[OF `wf_qtrs Q R` `s \<in> Tinf ?qr` ndef] by auto
      hence "?P'(next4 shp Q R s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
      have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
      have s: "?S 0 = fst(next4 shp Q R s)" unfolding concrete_seq.simps by simp
      have t: "?T 0 = fst(snd(next4 shp Q R s))" unfolding concrete_seq.simps by simp
      have sub: "?\<sigma> 0 = snd(snd(snd(next4 shp Q R s)))"
        unfolding concrete_seq.simps Let_def split_def by auto
      have nxt: "?I(Suc 0) = fst(snd(snd(next4 shp Q R s)))"
        unfolding concrete_seq.simps Let_def split_def by auto
      show ?case unfolding ini s t sub nxt using `?P'(next4 shp Q R s)` by auto
    next
      case (Suc i)
      let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
      from Suc have "?I (Suc i) \<in> Tinf ?qr" by simp
      have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)"
        using Tinf_lemma_DP_explicit_subst[OF `wf_qtrs Q R` `?I (Suc i) \<in> Tinf ?qr` ndef] by auto
      hence IH: "?P''(next4 shp Q R (?I(Suc i)))"
        unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
      have s: "?S(Suc i) = fst(next4 shp Q R (?I(Suc i)))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have t: "?T(Suc i) = fst(snd(next4 shp Q R (?I(Suc i))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 shp Q R (?I(Suc i)))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 shp Q R (?I(Suc i)))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      show ?case unfolding s t sub nxt using IH by auto
    qed
  qed
  from main have P: "\<forall>i. (?S i,?T i) \<in> DP shp (useful_rules Q R)" by simp
  from main have NF: "\<And>i. (?S i \<cdot> ?\<sigma> i) \<in> NF_terms ?Q" by auto
  from main have eq: "\<And>i. sharp_term shp (?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
  have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> ?Iqr^*"
  proof
    fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> ?Iqr^*"
    proof (cases i)
      case 0
      from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp_term shp (?I(Suc 0))" by best
      from main have "(sharp_term shp (?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> ?Iqr^*" by best
      thus ?thesis unfolding 0 unfolding first by simp
    next
      case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
    qed
  qed
  from P and R and NF have ichain: "ichain (initial_dpp shp Q R) ?S ?T ?\<sigma>" 
    by (auto simp: ichain.simps)
  have Tinf: "\<forall>i. ?I i \<in> Tinf ?qr"
  proof
    fix i show "?I i \<in> Tinf ?qr" using main by (induct i) (auto simp: `s \<in> Tinf ?qr`)
  qed
  show ?thesis
  proof (intro exI conjI, rule ichain, intro impI)
    assume ndef2: "\<forall> f n. defined (useful_rules Q R) (f,n) \<longrightarrow> \<not> defined (useful_rules Q R) (shp f, n)"
    have SN_t: "\<forall>i. SN_on ?Iqr {?T i \<cdot> ?\<sigma> i}"
    proof
      fix i
      from Tinf have Tinf: "?I (Suc i) \<in> Tinf ?qr" by best
      from `wf_qtrs Q R` have wwf: "wwf_qtrs Q R" unfolding wf_qtrs_def by auto
      from wwf_qtrs_imp_left_fun[OF wwf] have nvar: "\<forall> (l,r) \<in> R. is_Fun l" by auto
      from main have "(?S i, ?T i) \<in> DP shp (useful_rules Q R)" by auto
      then obtain f ts where "?T i = Fun (shp f) ts" and "defined (useful_rules Q R) (f,length ts)" unfolding DP_def by auto
      then obtain ts where id: "?T i \<cdot> ?\<sigma> i = Fun (shp f) ts" and d: "defined (useful_rules Q R) (f,length ts)"  by auto
      from eq id have eq: "sharp_term shp (?I(Suc i)) = Fun (shp f) ts" by simp
      show "SN_on ?Iqr {?T i \<cdot> ?\<sigma> i}" unfolding id 
      proof (rule SN_args_imp_SN[OF _ nvar ndef2[THEN spec, THEN spec, THEN mp[OF _ d]]])
        fix t
        assume t: "t \<in> set ts"
        from eq obtain g where "?I (Suc i) = Fun g ts" by (cases "?I (Suc i)", auto)
        from Tinf[unfolded this] t show "SN_on ?qr {t}" unfolding Tinf_def by auto
      qed
    qed
    from ichain SN_t
    show "min_ichain (initial_dpp shp Q R) ?S ?T ?\<sigma>" 
      unfolding initial_dpp.simps min_ichain.simps minimal_cond_def by auto
  qed
qed

hide_const next4 concrete_seq

corollary not_SN_imp_min_ichain:
  assumes prem: "wf_qtrs Q R" "\<not> SN (qrstep Q R)" 
  "\<And> f n ss. defined (useful_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  and ndef2: "\<And> f n. defined (useful_rules Q R) (f,n) \<Longrightarrow> \<not> defined (useful_rules Q R) (shp f,n)"
  shows "\<exists>s t \<sigma>. min_ichain (initial_dpp shp Q R) s t \<sigma>"
  using not_SN_imp_min_ichain_both[OF prem, of "\<lambda> x. x"] ndef2 by auto

corollary not_SN_imp_ichain:
  assumes prem: "wf_qtrs Q R" "\<not> SN (qrstep Q R)" 
  "\<And> f n ss. defined (useful_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  shows "\<exists>s t \<sigma>. ichain (initial_dpp shp Q R) s t \<sigma>"
  using not_SN_imp_min_ichain_both[OF prem, of "\<lambda> x. x"] by best


corollary not_SN_imp_ichain_rstep:
  assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" 
  shows "\<exists> s t \<sigma>. ichain (DP shp R, {}, {}, {}, R) s t \<sigma>"
proof -
  let ?DP = "DP shp (useful_rules {} R)"
  have U: "useful_rules {} R = R" unfolding useful_rules_def useful_rule_def by auto
  from wf U have wf: "wf_qtrs {} R" unfolding wf_qtrs_def wwf_qtrs_wf_trs by auto
  from nSN have nSN: "\<not> SN (qrstep {} R)" by auto
  have id: "initial_dpp shp {} R = (DP shp R, {}, {}, {}, R)" by (simp add: useful_rules_def useful_rule_def)
  from not_SN_imp_ichain[OF wf nSN, of shp, unfolded id U] show ?thesis by auto
qed

(* completeness lemma *)
lemma SN_imp_finite_dpp: fixes R :: "('f,'v)trs"
  assumes SN: "SN (qrstep Q R)"  
  and shp: "\<And> f n. defined R (f,n) \<Longrightarrow> \<not> defined R (shp f,n)"
  and unshp: "\<And> f n. defined R (f,n) \<Longrightarrow> unshp (shp f) = f"
  shows "finite_dpp (DP shp R,{},Q,{},R)"
proof (rule ccontr)
  let ?s = "sharp_term shp :: ('f,'v)term \<Rightarrow> ('f,'v)term"
  let ?u = "sharp_term unshp :: ('f,'v)term \<Rightarrow> ('f,'v)term"
  let ?r = "qrstep Q R \<union> op \<rhd>"
  assume "\<not> ?thesis"
  then obtain s t \<sigma> where "ichain (DP shp R,{},Q,{},R) s t \<sigma>" 
    unfolding finite_dpp_def by auto
  note ichain = this[unfolded ichain.simps]
  {
    fix l r x
    assume lr: "(l,r) \<in> R" and l: "l = Var x"
    from left_Var_imp_not_SN_qrstep[OF lr[unfolded l], of Q]
      SN have False unfolding SN_def by auto
  } note nvar = this
  hence varcond: "\<forall> (l,r) \<in> R. is_Fun l" by force
  {
    fix i
    from ichain have "(s i, t i) \<in> DP shp R" by auto
    from this[unfolded DP_def] obtain l r h us where s: "s i = ?s l"
      and t: "t i = ?s (Fun h us)" and lr: "(l,r) \<in> R" and ru: "r \<unrhd> Fun h us"
      and h: "defined R (h,length us)" by auto
    from nvar[OF lr]
    obtain f ls where l: "l = Fun f ls" by (cases l, auto)
    from lr[unfolded l] have f: "defined R (f, length ls)" unfolding defined_def by auto
    from ru obtain C where r: "r = C\<langle>Fun h us\<rangle>" by auto
    have si: "?u (s i \<cdot> \<sigma> i) =  l \<cdot> \<sigma> i" unfolding s l using unshp[OF f]
      by simp
    from ichain have "s i \<cdot> \<sigma> i \<in> NF_terms Q"
      by auto
    from NF_imp_subt_NF[OF this]
    have "\<forall> t \<in> set (args (s i \<cdot> \<sigma> i)). t \<in> NF_terms Q"
      unfolding NF_terms_args_conv .
    hence "\<forall> t \<in> set (args (l \<cdot> \<sigma> i)). t \<in> NF_terms Q" unfolding s l by simp
    hence NF: "\<forall> t \<lhd> l \<cdot> \<sigma> i. t \<in> NF_terms Q" unfolding NF_terms_args_conv .
    have step: "(l \<cdot> \<sigma> i, r \<cdot> \<sigma> i) \<in> qrstep Q R"
      by (rule qrstepI[OF _ lr, of "\<sigma> i" _ _ \<box>, OF NF], auto)
    have supt: "(r \<cdot> \<sigma> i, Fun h us \<cdot> \<sigma> i) \<in> (op \<rhd>)^=" using supteq_stable[OF ru, of "\<sigma> i"] unfolding supteq_supt_set_conv .
    from ichain have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q R)^*" by auto
    have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrqrstep Q R)^*"
    proof (rule qrsteps_imp_nrqrsteps[OF varcond ndef_useful_rules steps])
      show "\<not> defined R (the (root (t i \<cdot> \<sigma> i)))"
        unfolding t using shp[OF h] by simp
    qed
    from nrqrsteps_imp_sharp_qrsteps[OF steps, of unshp]
    have "(?u (t i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (qrstep Q R)^*" .
    hence steps: "(Fun h us \<cdot> \<sigma> i, ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (qrstep Q R)^*"
      unfolding t using unshp[OF h] by simp
    from si step supt steps have steps: "(?u (s i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in>
      qrstep Q R O (op \<rhd>)^= O (qrstep Q R)^*" by auto
    have "(?u (s i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> ?r^+"
      by (rule set_mp[OF _ steps], regexp)
  } note main = this
  obtain u where u: "u = (\<lambda> i. ?u (s i \<cdot> \<sigma> i))" by auto
  from main have "\<And> i. (u i, u (Suc i)) \<in> ?r^+" unfolding u .
  hence "\<not> SN (?r^+)" by auto
  hence "\<not> SN ?r" using SN_imp_SN_trancl[of ?r] by auto
  thus False using SN_imp_SN_union_supt[OF SN ctxt_closed_qrstep] ..
qed

end
