(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Trs_Impl
imports
  Trs
  Term_Impl
  Substitution
  Abstract_Rewriting_Impl
begin

type_synonym ('f, 'v) rules = "('f,'v) rule list"

fun
  rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "rewrite R s = concat (map (\<lambda> (l, r) . concat (map (\<lambda> p.
     (case match (s |_ p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>])
   ) (poss_list s)) ) R)"


lemma rewrite_complete_main:
  fixes l r::"('f,'v)term"
  assumes "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" 
  shows "\<exists> u \<in> set (rewrite R s). vars_term l \<supseteq> vars_term r \<longrightarrow> u = t" 
proof -
  let ?C = "ctxt_of_pos_term p s"
  from assms have lr: "(l,r) \<in> set R" 
    and p: "p \<in> poss s" 
    and s: "?C\<langle>l \<cdot> \<sigma>\<rangle> = s"
    and t: "?C\<langle>r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from s p have "?C\<langle>l \<cdot> \<sigma>\<rangle> = ?C\<langle>s|_p\<rangle>" by (simp add: ctxt_supt_id)
  hence "l \<cdot> \<sigma> = s|_p" by auto
  from this match_complete obtain \<delta> where 
    match: "match (s|_p) l = Some \<delta>" and 
      vars: "\<forall> x \<in> vars_term l. \<sigma> x = \<delta> x" by blast
  have l: "l \<cdot> \<sigma> = l \<cdot> \<delta>"
    by (rule term_subst_eq, insert vars, auto)
  obtain u where u: "u = ?C\<langle>r \<cdot> \<delta>\<rangle>" by auto
  {
    assume sub: "vars_term r \<subseteq> vars_term l"
    have "r \<cdot> \<sigma> = r \<cdot> \<delta>"
      by (rule term_subst_eq, insert sub vars, auto)
    with t have "u = t" unfolding u by simp
  } note ut = this
  have nearlyDone: "(l,r) \<in> set R \<and> p \<in> poss s \<and> u \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ ?C\<langle>r\<cdot>\<sigma>\<rangle>])" using lr p match unfolding u by auto 
  hence "u \<in> set ((\<lambda>(l, r). concat  (map 
     (\<lambda>p. case match (s|_p) l  of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>])
                          (poss_list s)))  (l,r))" (is "_ \<in> (?P (l,r))")
    by auto
  with lr have "\<exists> lr \<in> set R. u \<in> (?P lr)" by blast
  hence "u \<in> set (rewrite R s)" by auto
  thus ?thesis using ut by (auto simp del:rewrite.simps)
qed

lemma rewrite_complete:
  fixes l r::"('f,'v)term"
  assumes "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" 
  and "vars_term l \<supseteq> vars_term r"
  shows "t \<in> set (rewrite R s)"
  using rewrite_complete_main[OF assms(1)] using assms(2) by auto

lemma rewrite_sound: assumes "t \<in> set (rewrite R s)"
  shows "(s,t) \<in> rstep (set R)"
proof -
  from assms 
  have "\<exists> l r p.  (l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])" 
    by auto
  from this obtain l r p where one:
    "(l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])"
    by auto
  from this obtain \<sigma> where two: "match (s|_p) l = Some \<sigma> \<and> t \<in> {(ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>}"
    by (cases "match (s|_p) l", auto)
  hence match: "l \<cdot> \<sigma> = s|_p" using match_sound by auto 
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  thus "(s,t) \<in> rstep (set R)" unfolding rstep_iff_rstep_r_p_s by blast
qed

declare rewrite.simps[simp del]

definition first_rewrite :: "('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term option"
  where "first_rewrite R s \<equiv> case rewrite R s of Nil \<Rightarrow> None | Cons t _ \<Rightarrow> Some t"

definition compute_rstep_NF :: "('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term option"
  where "compute_rstep_NF R s \<equiv> compute_NF (first_rewrite R) s"

lemma compute_rstep_NF_sound: assumes res: "compute_rstep_NF R s = Some t"
  shows "(s, t) \<in> (rstep (set R))^*" using res[unfolded compute_rstep_NF_def]
proof (rule compute_NF_sound)
  fix s t
  assume "first_rewrite R s = Some t"
  from this[unfolded first_rewrite_def] obtain ts where "rewrite R s = t # ts"
    by (cases "rewrite R s", auto simp del: rewrite.simps)
  hence t: "t \<in> set (rewrite R s)" by simp
  from rewrite_sound[OF this] show "(s,t) \<in> rstep (set R)" .
qed

lemma compute_rstep_NF_complete: assumes res: "compute_rstep_NF R s = Some t"
  shows "t \<in> NF (rstep (set R))" using res[unfolded compute_rstep_NF_def]
proof (rule compute_NF_complete)
  fix s
  assume "first_rewrite R s = None"
  from this[unfolded first_rewrite_def] have empty: "rewrite R s = []" 
    by (cases "rewrite R s", auto simp del: rewrite.simps)
  show "s \<in> NF (rstep (set R))"
  proof(rule, rule)
    fix t
    assume "(s,t) \<in> rstep (set R)"
    from this[unfolded rstep_iff_rstep_r_p_s]
    obtain l r p \<sigma> where step: "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" by auto
    from rewrite_complete_main[OF step] arg_cong[OF empty, of set] show False 
      by auto
  qed
qed

definition
  check_join_NF ::
    "('f :: show, 'v :: show) rules \<Rightarrow> 
     ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_join_NF R s t \<equiv> case (compute_rstep_NF R s, compute_rstep_NF R t) of
      (Some s', Some t') \<Rightarrow> 
     check (s' = t') (
    ''the normal form '' +#+ shows s' +@+ shows '' of '' +@+ shows s +@+ shows '' differs from '' +@+ shows_nl +@+ shows ''the normal form '' +@+ shows t' +@+ shows '' of '' +@+ shows t)
       |  _ \<Rightarrow> error (shows ''strange error in normal form computation'')"

lemma check_join_NF_sound:
  assumes ok: "isOK (check_join_NF R s t)"
  shows "(s, t) \<in> join (rstep (set R))"
proof -
  note ok = ok[unfolded check_join_NF_def]
  from ok obtain s' where s: "compute_rstep_NF R s = Some s'" by force
  note ok = ok[unfolded s]
  from ok obtain t' where t: "compute_rstep_NF R t = Some t'" by force
  from ok[unfolded t] have id: "s' = t'" by simp
  note seq = compute_rstep_NF_sound
  from seq[OF s] seq[OF t]
  show ?thesis unfolding id by auto
qed

definition map_rules :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('g, 'v) rules" where
  "map_rules fg R = map (map_rule fg) R"

lemma map_rules_sound[simp]:
  "set (map_rules fg R) = map_trs fg (set R)"
  unfolding map_rules_def map_trs.simps by simp


subsection {* Output *}

fun shows_rule' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> ('f, 'v) rule \<Rightarrow> shows" where
  "shows_rule' fun var arr (l, r) =
    (shows_term' fun var l +@+ arr +#+ shows_term' fun var r)"

abbreviation "shows_rule \<equiv> shows_rule' shows shows '' -> ''"
abbreviation "shows_weak_rule \<equiv> shows_rule' shows shows '' ->= ''"

definition
  shows_rules' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows"
where
  "shows_rules' fun var arr trs \<equiv>
    shows_list_gen (shows_rule' fun var arr) [] [] [newline] [] trs +@+ shows_nl"

abbreviation "shows_rules \<equiv> shows_rules' shows shows '' -> ''"
abbreviation "shows_weak_rules \<equiv> shows_rules' shows shows '' ->= ''"

definition
  shows_trs' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> string \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows"
where
  "shows_trs' fun var name arr R \<equiv> name +#+ shows_nl +@+ shows_nl +@+ shows_rules' fun var arr R"

abbreviation "shows_trs \<equiv> shows_trs' shows shows ''rewrite system:'' '' -> ''"

definition
  vars_rule_list' :: "('f, 'v) rule \<Rightarrow> 'v list"
where
  "vars_rule_list' r \<equiv> vars_term_list' (fst r) @ vars_term_list' (snd r)"

lemma vars_rule_list'_sound[simp]:
  "set (vars_rule_list' r) = vars_rule r"
unfolding vars_rule_list'_def vars_rule_def by (cases r) simp

definition vars_rule_list :: "('f, 'v) rule \<Rightarrow> 'v list" where
  "vars_rule_list r \<equiv> remdups (vars_rule_list' r)"

lemma vars_rule_list_sound[simp]:
  "set (vars_rule_list r) = vars_rule r"
unfolding vars_rule_list_def by simp

definition vars_trs_list :: "('f, 'v) rule list \<Rightarrow> 'v list" where
  "vars_trs_list trs \<equiv> remdups (concat (map vars_rule_list' trs))"

lemma vars_trs_list_Nil[simp]: "vars_trs_list [] = []" unfolding vars_trs_list_def by simp

lemma vars_trs_list_sound[simp]:
  "set (vars_trs_list trs) = vars_trs (set trs)"
by (induct trs) (auto simp: vars_trs_def vars_trs_list_def)

lemmas vars_list_defs = vars_trs_list_def vars_rule_list_def

definition funs_rule_list' :: "('f, 'v) rule \<Rightarrow> 'f list" where
  "funs_rule_list' r \<equiv> funs_term_list' (fst r) @ funs_term_list' (snd r)"

lemma funs_rule_list'_sound[simp]:
  "set (funs_rule_list' r) = funs_rule r"
  unfolding funs_rule_list'_def funs_rule_def by (cases r) simp

definition funs_rule_list :: "('f, 'v) rule \<Rightarrow> 'f list" where
  "funs_rule_list r \<equiv> remdups (funs_rule_list' r)"

lemma funs_rule_list_sound[simp]:
  "set (funs_rule_list r) = funs_rule r"
  unfolding funs_rule_list_def by simp

definition funas_rule_list' :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list" where
  "funas_rule_list' r \<equiv> funas_term_list' (fst r) @ funas_term_list' (snd r)"

lemma funas_rule_list'_sound[simp]:
  "set (funas_rule_list' r) = funas_rule r"
  unfolding funas_rule_list'_def funas_rule_def by (cases r) simp

definition funas_rule_list :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list" where
  "funas_rule_list r \<equiv> remdups (funas_rule_list' r)"

lemma funas_rule_list_sound[simp]:
  "set (funas_rule_list r) = funas_rule r"
  unfolding funas_rule_list_def by simp

definition funas_args_rule_list' :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_rule_list' r \<equiv> funas_args_term_list' (fst r) @ funas_args_term_list' (snd r)"

lemma funas_args_rule'_sound[simp]:
  "set (funas_args_rule_list' r) = funas_args_rule r"
  unfolding funas_args_rule_list'_def funas_args_rule_def by (cases r) simp

definition funas_args_rule_list :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_rule_list r = remdups (funas_args_rule_list' r)"

lemma funas_args_rule_list_sound[simp]:
  "set (funas_args_rule_list r) = funas_args_rule r"
  unfolding funas_args_rule_list_def by simp

definition funs_trs_list :: "('f, 'v) rule list \<Rightarrow> 'f list" where
  "funs_trs_list R \<equiv> remdups (concat (map funs_rule_list' R))"

lemma funs_trs_list_sound[simp]:
  "set (funs_trs_list R) = funs_trs (set R)"
by (induct R) (auto simp: funs_trs_def funs_trs_list_def)

definition funas_root_rule_list :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list" where 
  "funas_root_rule_list r =
    remdups (option_to_list (root (fst r)) @ option_to_list (root (snd r)))"

lemma funas_root_rule_list_sound[simp]:
  "set (funas_root_rule_list rule) =
    Option.set (root (fst rule)) \<union> Option.set (root (snd rule))"
  by (cases rule) (simp_all add: funas_root_rule_list_def)

definition funas_root_rules_list :: "('f, 'v) rules \<Rightarrow> ('f \<times> nat) list" where
  "funas_root_rules_list r = remdups (concat (map funas_root_rule_list r))"

lemma funas_root_rules_list_sound[simp]:
  "set (funas_root_rules_list rs) = funas_root_trs (set rs)"
  unfolding funas_root_rules_list_def funas_defs by simp

definition funas_trs_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list" where
  "funas_trs_list R \<equiv> remdups (concat (map funas_rule_list' R))"

lemma funas_trs_list_sound[simp]:
  "set (funas_trs_list R) = funas_trs (set R)"
  by (induct R) (auto simp: funas_trs_def funas_trs_list_def)

definition funas_args_trs_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list" where
  "funas_args_trs_list R \<equiv> remdups (concat (map funas_args_rule_list' R))"

lemma funas_args_trs_list_sound[simp]:
  "set (funas_args_trs_list R) = funas_args_trs (set R)"
  by (induct R) (auto simp: funas_args_trs_def funas_args_trs_list_def)

lemmas funs_list_defs = funs_trs_list_def funs_rule_list_def
lemmas funas_list_defs = funas_trs_list_def funas_rule_list_def
lemmas funas_args_list_defs = funas_args_trs_list_def funas_args_rule_list_def

fun
  defs_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list"
where
  "defs_list [] = []" |
  "defs_list ((Fun f ss, r) # ts) = (let fs = defs_list ts; n = length ss in if (f,n) \<in> set fs then fs else (f,n) # fs)" |
  "defs_list ((Var x, r) # ts) = defs_list ts"

lemma defs_list_sound[simp]:
  "set (defs_list R) = {fn. defined (set R) fn}"
proof (induct R, simp add: defined_def)
  case (Cons lr R)
  show ?case
  proof (cases lr)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var x)
      with Pair Cons show ?thesis by (auto simp: defined_def)
    next
      case (Fun g ss)
      {
        fix f n
        from Pair Fun
        have "defined (set (lr # R)) (f, n) = defined (set ((Fun g ss,r) # R)) (f, n)" by simp
        also have "\<dots> = (\<exists>l ra. (l, ra) \<in> set ((Fun g ss, r) # R) \<and> root l = Some (f, n))"
          (is "_ = ?P (set ((Fun g ss,r) # R))") by (simp add: defined_def)
        also have "\<dots> = (?P (set R) \<or> ?P {(Fun g ss, r)})" by auto
        also have "\<dots> = ((f,n) \<in> set (defs_list R) \<or> ?P {(Fun g ss, r)})"  using Cons by (unfold defined_def, auto)
        also have "\<dots> = ((f,n) \<in> set (defs_list ((Fun g ss,r) # R)))" (is "?l = ?r")
        proof (cases "(g,n) \<in> set (defs_list R)")
	  case False
	  thus "?l = ?r" by (auto simp: Let_def)
        next
	  case True
	  thus "?l = ?r" by (simp add: Let_def, blast)
        qed
        finally have "((f,n) \<in> set (defs_list (lr # R))) = ((f,n) \<in> {fn. defined (set (lr # R)) fn})" using Fun Pair Cons by auto
      }	
      thus ?thesis by auto
    qed
  qed
qed

declare defs_list.simps [simp del]

lemma distinct_funas_trs_list:
  fixes trs::"('f,'v)rules" shows "distinct(funas_trs_list trs)"
  unfolding funas_trs_list_def by simp


definition
  check_left_linear_trs :: "('f::show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_left_linear_trs trs \<equiv>
     check_all (\<lambda>r. linear_term (fst r)) trs
       <+? (\<lambda> _. shows trs +@+ shows_nl +@+ ''is not left-linear'' +#+ shows_nl)"

lemma check_left_linear_trs[simp]: "isOK(check_left_linear_trs R) = (left_linear_trs (set R))"
  unfolding check_left_linear_trs_def left_linear_trs_def
  by auto

definition
  check_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_wf_trs R \<equiv> (
     check_allm (\<lambda>rule. do {
       check (is_Fun (fst rule)) (shows_string ''variable left-hand side in'');
       check_subseteq (vars_term_list (snd rule)) (vars_term_list (fst rule))
         <+? (\<lambda>x. ''free variable '' +#+ shows x
               +@+ shows_string '' in right-hand side of'')
     } <+? (\<lambda>s. s +@+ '' rule '' +#+ shows_rule rule +@+ shows_nl)
     ) R
   <+? (\<lambda>e. ''the TRS is not well-formed'' +#+ shows_nl +@+ e)
   )"

lemma check_wf_trs_conf[simp]:
  shows "isOK (check_wf_trs R) = wf_trs (set R)"
proof
  assume "isOK (check_wf_trs R)"
  hence "\<forall>(l, r)\<in>set R. is_Fun l" unfolding check_wf_trs_def by auto
  hence "\<forall>(l, r)\<in>set R. \<exists>f ts. l = Fun f ts" by best
  moreover have "\<forall>(l, r)\<in>set R. vars_term r \<subseteq> vars_term l"
  proof (intro ballI2)
    fix l r assume "(l, r) \<in> set R"
    from `isOK (check_wf_trs R)`
      have "isOK(check_allm (\<lambda>(l, r). check_subseteq (vars_term_list r) (vars_term_list l)) R)"
      unfolding check_wf_trs_def by auto
    hence "\<forall>(l, r)\<in>set R. isOK (check_subseteq (vars_term_list r) (vars_term_list l))"
      unfolding split_def by simp
    with `(l, r) \<in> set R` show "vars_term r \<subseteq> vars_term l" by auto
  qed
  ultimately show "wf_trs (set R)" unfolding wf_trs_def by auto
next
  assume "wf_trs (set R)" thus "isOK (check_wf_trs R)"
    unfolding wf_trs_def check_wf_trs_def by force
qed


definition
  check_not_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_not_wf_trs R \<equiv>
     check (\<not> isOK (check_wf_trs R)) (''The TRS is well formed'' +#+ shows_nl)"

lemma check_not_wf_trs:
  assumes "isOK(check_not_wf_trs R)"
  shows "\<not> SN (rstep (set R))"
proof -
  from assms have "\<not> wf_trs (set R)" unfolding check_not_wf_trs_def by auto
  with SN_rstep_imp_wf_trs show ?thesis by auto
qed

lemma [code]: "instance_rule lr st = (match_list (fst lr) [(fst lr, fst st), (snd lr, snd st)] \<noteq> None)" (is "?l = (match_list ?w ?list \<noteq> None)")
proof
  assume ?l
  then obtain \<sigma> where fst: "fst lr = fst st \<cdot> \<sigma>" and snd: "snd lr = snd st \<cdot> \<sigma>" unfolding instance_rule_def by auto
  have "\<And> t l. (t, l) \<in> set ?list \<Longrightarrow> l \<cdot> \<sigma> = t" by (auto simp del: match_list.simps simp: fst snd)
  from match_list_complete[of ?list \<sigma>, OF this]
  obtain \<sigma>' where "match_list ?w ?list = Some \<sigma>'" by blast
  thus "match_list ?w ?list \<noteq> None" by simp
next
  assume "match_list ?w ?list \<noteq> None"
  then obtain \<sigma> where "match_list ?w ?list = Some \<sigma>" by (cases "match_list ?w ?list", auto)
  from match_list_sound[OF this] have "fst lr = fst st \<cdot> \<sigma>" and "snd lr = snd st \<cdot> \<sigma>" by (force+)
  thus ?l unfolding instance_rule_def by auto
qed


definition check_CS_subseteq :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule check"
where "check_CS_subseteq R S \<equiv> check_allm (\<lambda> (l,r). check (list_ex (instance_rule (l,r)) S) (l,r)) R"

lemma check_CS_subseteq[simp]: "isOK(check_CS_subseteq R S) = (CS (set R) \<subseteq> CS (set S))" (is "?l = ?r")
proof
  assume ?l
  show ?r
  proof
    fix x y
    assume "(x,y) \<in> CS (set R)"    
    thus "(x,y) \<in> CS (set S)"
    proof (induct)
      case (subst_closureI s t \<sigma>)
      with `?l`[unfolded check_CS_subseteq_def]
      obtain l r \<delta> where lr: "(l,r) \<in> set S" and s: "s = l \<cdot> \<delta>" and t: "t = r \<cdot> \<delta>"
        unfolding list_ex_iff by (force simp: instance_rule_def)
      show ?case unfolding s t
        using subst_closure.subst_closureI[OF lr, of "\<delta> \<circ>s \<sigma>"]
        by auto
    qed
  qed
next
  assume ?r
  show ?l
    unfolding check_CS_subseteq_def
  proof (simp, intro ballI)
    fix lr
    assume mem: "lr \<in> set R"
    obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
    with mem have "(l,r) \<in> CS (set R)" using subset_subst_closure by auto
    with `?r` have "(l,r) \<in> CS (set S)" by auto
    thus "list_ex (instance_rule lr) S" unfolding lr
    proof (induct)
      case (subst_closureI s t \<sigma>)
      thus ?case unfolding list_ex_iff instance_rule_def by force
    qed
  qed
qed


instantiation pos :: "show" begin
  fun
    shows_pos :: "pos \<Rightarrow> shows"
  where
    "shows_pos (i <# \<epsilon>) = shows (Suc i)" |
    "shows_pos (i <# ps) = (shows (Suc i) +@+ shows_pos ps)" |
    "shows_pos \<epsilon> = shows_string ''epsilon''"
  definition "shows_prec (d::nat) (p::pos) = shows_pos p"
  definition "shows_list_pos (ps::pos list) = shows_list_aux shows ps"

  lemma assoc_pos: "shows_prec d (p::pos) r @ s = shows_prec d p (r @ s)"
  unfolding shows_prec_pos_def
  proof (induct p)
    case (PCons i ps) thus ?case by (cases ps) simp_all
  qed simp

  lemma assoc_pos_elt: "\<forall>(p::pos)\<in>set ps. shows p r @ s = shows p (r @ s)"
  using assoc_pos by auto

  instance by
    (intro_classes, rule assoc_pos, unfold shows_list_pos_def)
    (rule shows_list_aux_assoc[OF assoc_pos_elt])
end


type_synonym ('f,'v)rule_map = "(('f \<times> nat) \<Rightarrow> ('f,'v)rules)option"

fun computeRuleMapH :: "('f,'v)rules \<Rightarrow> (('f \<times> nat) \<times> ('f,'v)rules)list option"
where "computeRuleMapH [] = Some []"
    | "computeRuleMapH ((Fun f ts,r) # rules) = (let n = length ts in case computeRuleMapH rules of None \<Rightarrow> None | Some rm \<Rightarrow> 
            (case extract (\<lambda> (fa,rls). fa = (f,n)) rm of 
                None \<Rightarrow> Some (((f,n), [(Fun f ts,r)]) # rm)
              | Some (bef,(fa,rls),aft) \<Rightarrow>  Some  ((fa,(Fun f ts,r) # rls) # bef @ aft)))"
    | "computeRuleMapH ((Var _, _) # rules) = None"

definition computeRuleMap :: "('f,'v)rules \<Rightarrow> ('f,'v)rule_map"
 where "computeRuleMap rls \<equiv> case computeRuleMapH rls of
           None \<Rightarrow> None
         | Some rm \<Rightarrow> Some (\<lambda> f. case lookup f rm of
               None \<Rightarrow> []
            | Some rls \<Rightarrow> rls)"

lemma computeRuleMapHSound2: "(computeRuleMapH R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
proof (induct R, simp)
  case (Cons rule rules)
  show ?case 
  proof (cases rule)
    case (Pair l r)
    show ?thesis 
    proof (cases l, simp add: Pair Cons)
      case (Fun f ts)
      show ?thesis
        using Pair Cons
      proof (cases "computeRuleMapH rules", simp add: Pair Fun Let_def)
        case (Some rm) note oSome = this
        let ?e = "extract (\<lambda> (fa,rls). fa = (f,length ts)) rm"
        from Some Fun Pair Cons show ?thesis 
        proof (cases ?e, simp)
          case (Some res)
          then obtain bef aft g rls where "?e = Some (bef, (g,rls), aft)" by (cases res, force) 
          with extract_Some[OF this] have rm: "rm = bef @ ((f, length ts),rls) # aft" and e: "?e = Some (bef, ((f,length ts),rls), aft)"
            by auto
          show ?thesis using Cons
            by (simp add: Pair Fun Let_def oSome e)
        qed
      qed
    qed
  qed
qed

lemma computeRuleMapSound2: "(computeRuleMap R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
unfolding computeRuleMap_def
by (simp only: computeRuleMapHSound2[symmetric], cases "computeRuleMapH R", auto)


lemma computeRuleMapHSound: assumes "computeRuleMapH R = Some rm"
  shows "(\<lambda> (f,rls). (f,set rls)) ` set rm = {((f,n),{(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}) | f n. {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)} \<noteq> {}} \<and> distinct_eq (\<lambda> (f,rls) (g,rls'). f = g) rm"
using assms
proof (induct R arbitrary: rm, force)
  case (Cons rule rules)
  let ?setl = "\<lambda> rm. (\<lambda> (f,rls). (f,set rls)) ` set rm"
  let ?setr = "\<lambda> R. {((f,n),{(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}) | f n. {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)} \<noteq> {}}"
  show ?case
  proof (cases rule)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var v)
      with Cons Pair show ?thesis by simp
    next
      case (Fun f ts)
      with Cons Pair show ?thesis 
      proof (cases "computeRuleMapH rules", simp)
        case (Some rrm) note oSome = this
        let ?dis = "distinct_eq (\<lambda> (f,rls) (g,rls'). f = g)"
        from Cons(1)[OF Some] have drrm: "?dis rrm" and srrm: "?setl rrm = ?setr rules" by auto
        show ?thesis
        proof (cases "extract (\<lambda> (fa,rls). fa = (f,length ts)) rrm")
          case None
          let ?e = "((f,length ts), [(Fun f ts,r)])"
          let ?e' = "((f,length ts), {(Fun f ts,r)})"
          from None Cons(2) have rm: "rm = ?e # rrm" by (simp add: Fun Pair Some None)
          from extract_None[OF None] have rrm: "\<And> g n rl. ((g,n),rl) \<in> set rrm \<Longrightarrow> (f,length ts) \<noteq> (g,n)" by auto
          hence rrm': "\<And> g n rl. ((g,n),rl) \<in> ?setr rules \<Longrightarrow> (f,length ts) \<noteq> (g,n)" by (simp only: srrm[symmetric], auto)
          hence  id: " {(Fun f ts, r)} = {(l, ra). (l = Fun f ts \<and> ra = r \<or> (l, ra) \<in> set rules) \<and> root l = Some (f, length ts)}"  by force
          from rrm have dis: "?dis rm" 
            by (simp add: rm drrm, auto)
          have "?setl rm = insert ?e' (?setl rrm)" by (simp add: rm)
          also have "\<dots> = insert ?e' (?setr rules)" by (simp add: srrm)
          also have "\<dots> = ?setr ( (Fun f ts,r) # rules)" 
          proof (rule set_eqI, clarify)
            fix g n rls
            show "(((g,n),rls) \<in> insert ?e' (?setr rules)) = (((g,n),rls) \<in> ?setr ((Fun f ts,r) # rules))"
            proof (cases "(g,n) = (f,length ts)")
              case False
              hence "(((g,n),rls) \<in> insert ?e' (?setr rules)) = (((g,n),rls) \<in> ?setr rules)" by auto
              also have "\<dots> = (((g,n),rls) \<in> ?setr ((Fun f ts,r) # rules))" using False by auto
              finally show ?thesis .
            next
              case True note oTrue = this
              show ?thesis 
              proof (cases "rls = {(Fun f ts, r)}")
                case True
                with oTrue show ?thesis by (simp add: id, force)
              next
                case False
                show ?thesis using rrm'[of g n rls] True False by (simp add: False True id, auto)
              qed
            qed
          qed
          finally show ?thesis
            by (simp only: dis drrm, simp add: Pair Fun)
        next
          case (Some res)
          obtain bef fg rls aft where "res = (bef,(fg,rls),aft)" by (cases res, force)
          from extract_Some[OF Some[simplified this]] Some[simplified this] have rrm: "rrm = bef @ ((f,length ts), rls) # aft" 
            and e: "extract (\<lambda> (fa, rls). fa = (f,length ts)) rrm = Some (bef, ((f,length ts),rls),aft)" by auto
          let ?e = "((f,length ts), (Fun f ts,r) # rls)"
          let ?e' = "((f,length ts), insert (Fun f ts,r) (set rls))"
          have "((f,length ts),set rls) \<in> ?setl rrm" unfolding rrm by auto
          hence rls: "set rls = {(l, r) |l r. (l, r) \<in> set rules \<and> root l = Some (f, length ts)}" using Cons(1)[OF oSome] by auto
          obtain ba where ba: "ba = bef @ aft" by auto
          from Cons(2) e ba have rm: "rm = ?e # ba" by (simp add: Fun Pair oSome e)
          from drrm[simplified rrm] have dis: "?dis ba"  unfolding distinct_eq_append ba by auto
          from drrm[simplified rrm] have dis: "?dis rm" unfolding rm distinct_eq_append ba by (auto simp: dis[simplified ba])
          from drrm[simplified rrm distinct_eq_append] 
          have diff: "(\<forall>x\<in>set ba. \<not> (\<lambda>(g, rls). (f,length ts) = g) x)" by (auto simp: ba)
          have "?setl [((f, length ts),rls)] \<union> ?setl ba = ?setl rrm" using rrm ba by auto
          also have "\<dots> = ?setr rules" by (rule srrm)
          finally have id: "?setl [((f, length ts),rls)] \<union> ?setl ba = ?setr rules" .          
          have "?setl rm = insert ?e' (?setl ba)" by (simp add: rm)
          also have "\<dots> = ?setr ((Fun f ts,r) # rules)" 
          proof (rule set_eqI, clarify)
            fix g n rl 
            show "(((g,n),rl) \<in> insert ?e' (?setl ba)) = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))"
            proof (cases "(g,n) = (f,length ts)")
              case False 
              hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = (((g,n),rl) \<in> ?setl ba)" by auto
              also have "\<dots> = (((g,n),rl) \<in> ?setr rules)" using False by (simp only: id[symmetric], auto)
              also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" using False by auto
              finally show ?thesis .
            next
              case True note oTrue = this
              show ?thesis 
              proof (cases "rl = insert (Fun f ts, r) (set rls)")
                case True
                hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = True" using oTrue by auto
                also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" unfolding True rls using oTrue by force
                finally show ?thesis .
              next
                case False
                hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = False" using diff by (simp add: True, auto)
                also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" 
                proof (rule ccontr)
                  assume "\<not> ?thesis"
                  with True have "((f,length ts),rl) \<in> ?setr ((Fun f ts,r) # rules)" by simp
                  hence "rl = {(l, ra) |l ra. (l, ra) \<in> set ((Fun f ts, r) # rules) \<and> root l = Some (f, length ts)}" by simp
                  with False rls show False by auto
                qed                              
                finally show ?thesis .
              qed
            qed
          qed
          also have "\<dots> = ?setr (rule # rules)" by (simp add: Pair Fun)
          finally show ?thesis by (simp add: dis)
        qed
      qed
    qed
  qed
qed

lemma computeRuleMapSound:
  assumes "computeRuleMap R = Some rm" 
  shows "(set (rm (f,n))) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
proof (cases "computeRuleMapH R")
  case None
  thus ?thesis using assms unfolding computeRuleMap_def by auto
next
  case (Some rrm)
  note rrm = computeRuleMapHSound[OF this] 
  note rm = assms[unfolded computeRuleMap_def, simplified Some, simplified, symmetric]
  show ?thesis
  proof (cases "lookup (f,n) rrm")
    case (Some rls)
    from lookup_sound[OF this] have "((f,n),set rls) \<in> (\<lambda> (f,rls). (f, set rls)) ` set rrm"
      by auto
    hence "set rls = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
      by (simp only: rrm, simp) 
    thus ?thesis by (simp add: rm Some)
  next
    case None
    have id: "{(l, r) |l r. (l, r) \<in> set R \<and> root l = Some (f, n)} = {}" (is "?set = {}")
    proof (rule ccontr)
      assume "\<not> ?thesis"
      then obtain l r where "(l, r) \<in> set R \<and> root l = Some (f, n)" by auto
      with rrm have "(((f,n), ?set)) \<in> (\<lambda> (f,rls). (f, set rls)) ` set rrm" by auto
      with lookup_complete[OF None] show False by auto
    qed
    thus ?thesis
      by (simp only: rm None id, auto)
  qed
qed
    
    
  
lemma computeRuleMap_left_vars:
  shows "(computeRuleMap R \<noteq> None) = (\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x)"
proof (cases "computeRuleMap R")
  case None
  from None computeRuleMapSound2 have "\<exists> (l,r) \<in> set R. root l = None" by auto
  from this obtain l r where "(l,r) \<in> set R \<and> root l = None" by auto
  from this have "(l,r) \<in> set R \<and> \<not> (\<forall> x. fst (l,r) \<noteq> Var x)" by (cases l, auto)
  with None show ?thesis by blast
next
  case (Some rm)
  hence left: "computeRuleMap R \<noteq> None" by auto
  from Some computeRuleMapSound2 have "\<forall> (l,r) \<in> set R. root l \<noteq> None" by force
  hence "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" by auto
  with left show ?thesis by blast
qed

lemma computeRuleMap_defs: fixes R :: "('f,'v)rules"
  assumes "computeRuleMap R = Some rm"
  shows "(rm (f,n) = []) = ((f,n) \<notin> defs (set R))"
proof -
  from assms computeRuleMapSound have rm: "\<And>(f::'f) n. set (rm (f,n)) = {(l,r) | l r. (l, r) \<in> set R \<and> root l = Some (f, n)}" by force
  show ?thesis
  proof (cases "rm (f,n)")
    case Nil
    with rm have "(f,n) \<notin> defs (set R)" unfolding defs_def defined_def by force
    with Nil show ?thesis by blast
  next
    case (Cons lr RR)
    hence left: "rm (f,n) \<noteq> []" by auto
    from Cons rm[where f = f and n = n] have "(f,n) \<in> defs (set R)" unfolding defs_def defined_def by (cases lr, force)
    with left show ?thesis by blast
  qed
qed



lemma computeRuleMap_None_not_SN:
  assumes "computeRuleMap R = None"
  shows "\<not> SN_on (rstep (set R)) {t}"
proof -
  from assms computeRuleMap_left_vars[of R] obtain x r where "(Var x,r) \<in> set R" by auto
  from left_var_imp_not_SN[OF this] show ?thesis .
qed


definition reverse_rules :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rules" where
  "reverse_rules rs \<equiv> map (\<lambda>lr. (snd lr, fst lr)) rs"

lemma reverse_rules[simp]: "set (reverse_rules R) = (set R)^-1" unfolding reverse_rules_def by force

definition
  map_rules_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('g, 'v) rules"
where
  "map_rules_wa fg R \<equiv> map (\<lambda>(l, r). (map_term_wa fg l, map_term_wa fg r)) R"

lemma map_rules_wa: "set (map_rules_wa fg R) = map_trs_wa fg (set R)"
  unfolding map_rules_wa_def map_trs_wa_def by auto


end
